(*	$Id: Calendar.Mod,v 1.1 1997/02/07 07:45:32 oberon1 Exp $	*)
MODULE Calendar;

(*
    Calendar - routines to manipulate dates/times.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT CI:=CharClass, T:=Time, SC:=SysClock, JD:=JulianDay, S:=Strings, 
       IntStr, L:=Locales, Out;

CONST 
  (* days of the week *)
  Sunday * = 0; Monday * = 1; Tuesday * = 2; Wednesday * = 3;
  Thursday * = 4; Friday * = 5; Saturday * = 6;
  
  (* months of the year *)
  January * = 1; February * = 2; March * = 3; April * = 4;
  May * = 5; June * = 6; July * = 7; August * = 8;
  September * = 9; October * = 10; November * = 11; December * = 12;

  (* when debugging this module *)
  DEBUG = FALSE;

TYPE
  (* local calendar definition *)
  Calendar* = RECORD (SC.DateTime)
  END;


(* ------------------------------------------------------------- *)
(* Internal conversion functions *)

PROCEDURE IsValidTime (h, m, s: SHORTINT; f: INTEGER) : BOOLEAN;
BEGIN
  RETURN (h>=0)&(h<24)&(m>=0)&(m<60)&(s>=0)&(s<60)&(f>=0)&(f<=SC.maxSecondParts)
END IsValidTime;

PROCEDURE TimeToMSec (h, m, s: SHORTINT; f: INTEGER) : LONGINT;
(* Converts the time to the number of milliseconds since 00:00 
   local time or 0 on invalid input times. *)
BEGIN
  IF IsValidTime(h, m, s, f) THEN 
    RETURN ((h*60+m)*60+s)*1000+f
  ELSE RETURN 0
  END
END TimeToMSec;

PROCEDURE MSecToTime (ms: LONGINT; VAR h, m, s: SHORTINT; VAR f: INTEGER);
(* Converts the milliseconds since 00:00 to a local time.
   Pre: 0<=m<8.64E7; 
   Post: h, m, s are set to the 24-hour time; f is milliseconds < 1000
*)
BEGIN
  IF (ms>=0) & (ms<T.msecPerDay) THEN 
    f:=SHORT(ms MOD 1000); ms:=ms DIV 1000; 
    s:=SHORT(SHORT(ms MOD 60)); ms:=ms DIV 60;
    m:=SHORT(SHORT(ms MOD 60));
    h:=SHORT(SHORT(ms DIV 60))
  ELSE f:=0; h:=0; m:=0; s:=0
  END
END MSecToTime;


(* ------------------------------------------------------------- *)
(* Various initialization functions *)

PROCEDURE Init* (VAR c: Calendar; d, m: SHORTINT; y: INTEGER; 
		 h, min, s: SHORTINT);
(* Initialize the calendar `c' with the date from `d' days, `m' months,
   `y' years; and the local time from `h' hours, `min' minutes, and
   `s' seconds.  Where:

       y > 0; 1<=m<=12; 1<=d<=31; 0<=h<24; 0<=min<60; 0<=s<60 
		
   The time zone and daylight savings information are obtained
   from the current system time (via SysClock). *)
BEGIN
  (* obtain the time zone and dst mode *)
  SC.GetClock(c);

  (* initialize the calendar *)
  c.day:=d; c.month:=m; c.year:=y;
  c.fractions:=0; c.second:=s; c.minute:=min; c.hour:=h
END Init;

PROCEDURE InitWithZoneDST (VAR c: Calendar; s: T.TimeStamp;
                           z: INTEGER; dst: SHORTINT);
VAR
  ms, days: LONGINT;                          
BEGIN  
  (* adjust for the time zone *)
  IF z<=SC.unknownZone THEN ms:=s.msecs 
  ELSE ms:=s.msecs-z*T.msecPerMin
  END; 

  (* check for overflows *)
  days:=s.days;
  IF ms>T.msecPerDay THEN INC(days); DEC(ms, T.msecPerDay)
  ELSIF ms<0 THEN DEC(days); INC(ms, T.msecPerDay) 
  END;

  (* convert to a local date/time *)
  MSecToTime(ms, c.hour, c.minute, c.second, c.fractions);
  JD.DaysToDate(days, c.day, c.month, c.year);
  c.zone:=z; c.summerTimeFlag:=dst
END InitWithZoneDST;

PROCEDURE InitFromTimeStamp * (VAR c: Calendar; s: T.TimeStamp);
(* Initialize the calendar from the time stamp `s'.  The time
   zone and daylight savings information is obtained from the
   current system time (via SysClock). *)
BEGIN
  (* obtain the time zone & dst information *)
  SC.GetClock(c);
  
  (* fill in from the time stamp *)
  InitWithZoneDST(c, s, c.zone, c.summerTimeFlag)
END InitFromTimeStamp;


(* ------------------------------------------------------------- *)
(* DateTime/TimeStamp conversion functions *)

PROCEDURE (VAR c: Calendar) ToTimeStamp * (VAR t: T.TimeStamp);
(* Converts the calendar date in `c' to a TimeStamp in `t'. *)
VAR
  days, ms: LONGINT;
BEGIN
  (* adjust for the time zone *)
  days:=JD.DateToDays(c.day, c.month, c.year);
  IF c.zone<=SC.unknownZone THEN 
    ms:=TimeToMSec(c.hour, c.minute, c.second, c.fractions)
  ELSE 
    ms:=TimeToMSec(c.hour, c.minute, c.second, c.fractions)+c.zone*T.msecPerMin
  END;

  (* check for overflows *)
  IF ms<0 THEN DEC(days); INC(ms, T.msecPerDay) 
  ELSIF ms>T.msecPerDay THEN INC(days); DEC(ms, T.msecPerDay)
  END;

  (* initialize the TimeStamp *)
  T.InitTimeStamp(t, days, ms) 
END ToTimeStamp;


(* ------------------------------------------------------------- *)
(* Date manipulation functions *)

PROCEDURE ^ (VAR c: Calendar) DaysPerMonth * () : SHORTINT; (* forward decl *)

PROCEDURE (VAR c: Calendar) Inc * (days, secs: LONGINT);
(* Add `days' days to the date `c'.  `days' and `secs' can be 
   negative. *)
VAR
  int: T.Interval; t: T.TimeStamp;
BEGIN
  T.InitInterval(int, days, T.msecPerSec*secs);   (* create interval *)
  c.ToTimeStamp(t);                               (* convert to TimeStamp *)
  t.Add(int);                                     (* and add the interval *)           
  InitWithZoneDST(c, t, c.zone, c.summerTimeFlag) (* convert to calendar *)
END Inc;

PROCEDURE (VAR c: Calendar) IncByMonth * (mon: LONGINT);
(* Advance the calendar `c' by `mon' months.  `mon' can be
   negative.  The day will be unchanged or shifted to the
   last day of the month if the new month has fewer days
   than the current month and the day falls on a missing
   day. `c.zone' should be valid. The time remains unchanged. *)
VAR
  years: INTEGER; day: SHORTINT;
BEGIN
  mon:=c.month-1+mon;
  IF mon<0 THEN years:=-SHORT((-mon) DIV 12)-1
  ELSE years:=SHORT(mon DIV 12)
  END;
  c.month:=SHORT(SHORT(mon-years*12+1)); INC(c.year, years);
  day:=c.DaysPerMonth();
  IF c.day>day THEN c.day:=day END
END IncByMonth;

PROCEDURE (VAR c: Calendar) IncByYear * (year: LONGINT);
(* Advance the calendar `c' by `year' years.  `year' can be
   negative.  The day will be unchanged or shifted to the
   last day of the month if the new month has fewer days
   than the current month and the day falls on a missing
   day. The time remains unchanged. *)
BEGIN
  c.IncByMonth(year*12)
END IncByYear;


(* ------------------------------------------------------------- *)
(* Date selector functions *)

PROCEDURE (VAR c: Calendar) DayOfWeek * () : SHORTINT;
(* Returns the day of week of `days' where
       0 - Sunday
       1 - Monday
       .
       .
       .
       6 - Saturday 
 *)
VAR
  t: T.TimeStamp;
BEGIN
  c.ToTimeStamp(t);  (* convert to GMT days since startMJD is also GMT *)
  RETURN SHORT(SHORT(ENTIER(t.days+JD.startMJD+1.5) MOD 7))
END DayOfWeek;

PROCEDURE (VAR c: Calendar) DaysPerMonth * () : SHORTINT;
VAR
  d, m: SHORTINT; y: INTEGER;
BEGIN
  CASE c.month OF
  |  1, 3, 5, 7, 8, 10, 12: RETURN 31
  |  2: JD.DaysToDate(JD.DateToDays(28, 2, c.year)+1, d, m, y);
        IF m#February THEN RETURN 28 ELSE RETURN 29 END
  |  ELSE RETURN 30
  END
END DaysPerMonth;

PROCEDURE (VAR c: Calendar) FindDate * (Week, DOW: SHORTINT; Start: Calendar);
(* Sets `c' to the first date which is a certain `DOW' where 0-Sunday, etc
   and falling in the nth week where n is given by `Week' and starting with the
   `Start' date.  When `Week' is negative, the `DOW' falling in the nth
   last week beginning at the date in `Start' will be returned.
   Valid ranges of inputs are: Week#0, 0<=DOW<7.     
   For example: to find the 3rd Sunday in December 1996, Week=3, DOW=0, 
   Start=1 Dec 1996. The returned date `c' will be 15 Dec 1996.  Also
   to find the 2nd last Monday in December 1996, the arguments become
   Week=-2, DOW=1, Start=31 Dec 1996.  The returned `c' will be 23 Dec 1996. *)
VAR 
  inc: SHORTINT;
BEGIN
  IF Week<0 THEN Week:=ABS(Week); inc:=-1 ELSE inc:=1 END;
  LOOP   
    IF c.DayOfWeek()=DOW THEN c.Inc((Week-1)*7*inc, 0); EXIT END;
    c.Inc(inc, 0)
  END
END FindDate;

PROCEDURE (VAR c: Calendar) WeekNumber * (startday: SHORTINT) : INTEGER;
(* Return the week number this week belongs to starting the count at `startday'
   which takes values of 0-Sunday, 1-Monday, etc.  The first week of a month
   is recognized as having 4 or more days where each week begins on `startday'. *)
VAR
  first, last, t: T.TimeStamp;
  
  PROCEDURE StartDate (year: INTEGER; VAR d: T.TimeStamp);
  VAR
    lc: Calendar;
  BEGIN
    Init(lc, 1, January, year, 0, 0, 0);
    lc.FindDate(1, startday, lc);
    IF lc.day>4 THEN lc.Inc(-7, 0) END;  (* start in prev. week *)
    lc.ToTimeStamp(d)
  END StartDate;
  
BEGIN  
  (* first determine the date corresponding to `startday' *)
  StartDate(c.year, first);
  
  (* also find the start date for the next year *)
  StartDate(c.year+1, last);
  
  (* how many weeks between `Date' and `first' *) 
  c.ToTimeStamp(t);
  IF t.days<first.days THEN
    (* count the weeks from previous year's start *)
    StartDate(c.year-1, first);
    RETURN SHORT((t.days-first.days) DIV 7)+1
  ELSIF t.days>=last.days THEN 
    (* Date is in the start of the following year *)
    RETURN 1
  ELSE
    (* Date is in this year *)
    RETURN SHORT((t.days-first.days) DIV 7)+1   
  END
END WeekNumber;

PROCEDURE (VAR c: Calendar) DayOfYear * (): INTEGER;
(* Return day of year where Jan 1st would be 1 *)
VAR ndate: Calendar; fs, cs: T.TimeStamp;
BEGIN
  ndate:=c; ndate.day:=1; ndate.month:=January;
  ndate.ToTimeStamp(fs); c.ToTimeStamp(cs);
  RETURN SHORT(cs.days - fs.days + 1);
END DayOfYear;


(* ------------------------------------------------------------- *)
(* Miscellaneous functions *)

PROCEDURE (VAR c: Calendar) MoonPhase * () : SHORTINT;
(* Returns the phase of the moon from 0 to 29 days for the current date
   in `c' where the phase can be obtained as follows:

             0 new moon 
             7 first quarter
            14 full moon
            22 last quarter

   This algorithm is based on one published in Sky & Telescope, April
   1994 by Bradley Schaefer and should be accurate for any date
   supported for the current calendar. Note: This algorithm does
   not take into account the observer's longitude so it may differ
   from the local astronomical moon phase by a day or so. *)
CONST
  MeanLunarCycle = 29.530588853D0;  (* length of lunar cycle *)
  JulianRefDate  = 2451550.1D0;     (* new moon in Julian days *)
VAR
  jd, phase: LONGREAL;
BEGIN
  (* compute the Julian days at 1200 GMT *)
  jd:=JD.DateToJD(c.day, c.month, c.year)+0.5;

  (* compute the moon's age (phase) *)
  phase:=(jd-JulianRefDate)/MeanLunarCycle;

  (* normalize the phase *)
  phase:=phase-ENTIER(phase);
  IF phase<0 THEN phase:=phase+1 END;
  RETURN SHORT(SHORT(ENTIER(phase*MeanLunarCycle)))
END MoonPhase;

PROCEDURE (VAR c: Calendar) PaschalFullMoon * (year: INTEGER);
(* This function sets the date of the Paschal Full Moon in `c'
   for the given `year'.  This function is valid for any date 
   between 1583 and 4099 (Gregorian calendar).  The Paschal full
   moon is used to compute various religious dates such as Easter,
   Passover, Good Friday, etc.  This full moon IS NOT the same as
   the astronomical full moon and can differ by a few days.  The
   chief advantage of the Paschal full moon is that it is independent
   of the observer's time zone and the exact date of the spring 
   equinox. Algorithm is due to John Hisco (jhisco@nexus.edu.au) *)
VAR
  a, b, d: INTEGER;
BEGIN
  a:=year DIV 100; b:=year MOD 19;

  (* determine the Paschal full moon date *)
  d:=(a-15) DIV 2 + 202 - 11*b;
  IF a>38 THEN INC(d, 2) ELSIF a>26 THEN INC(d) END;
  IF (a=21) OR (a=24) OR (a=25) OR (a>=33) & (a<=37) THEN INC(d) END;
  d:=d MOD 30 + 21;
  IF (d=29) OR (d=28) & (b>10) THEN INC(d) END;
 
  (* intialize the date *)
  IF d>31 THEN Init(c, SHORT(d-31), 4, year, 0, 0, 0)
  ELSE Init(c, SHORT(d), 3, year, 0, 0, 0)
  END 
END PaschalFullMoon;

PROCEDURE (VAR c: Calendar) DayOfWeekStr* (len: INTEGER; VAR Str: ARRAY OF CHAR);
(* Returns a localized day of week string for the date in `d' with a length
   of `len' characters where Sunday=0, Monday=1, etc. *)
VAR
  Day: ARRAY 32 OF CHAR;
BEGIN
  ASSERT(L.GetStr(L.daysOfWeek+c.DayOfWeek(), Day)); (* localized day of week *)
  IF len=0 THEN COPY(Day, Str)
  ELSE S.Extract(Day, 0, len, Str)
  END
END DayOfWeekStr;

PROCEDURE (VAR c: Calendar) MonthStr * (Short: BOOLEAN; VAR Str: ARRAY OF CHAR);
(* Returns a localized month string for the date in `c' using three
   characters when `Short' is TRUE and the full month otherwise. *)
VAR
  Month: ARRAY 32 OF CHAR;
BEGIN
  ASSERT(L.GetStr(L.months+c.month-1, Month));
  IF Short THEN S.Extract(Month, 0, 3, Str)
  ELSE COPY(Month, Str)
  END
END MonthStr;


(* ------------------------------------------------------------- *)
(* local string conversion assistance functions *)

PROCEDURE ExtractNumber (VAR str: ARRAY OF CHAR; VAR num: INTEGER; digits: INTEGER);
BEGIN
  num:=0;
  WHILE CI.IsNumeric(str[0]) & (digits>0) DO 
    num:=10*num+ORD(str[0])-ORD('0'); S.Delete(str, 0, 1); DEC(digits) 
  END  
END ExtractNumber;

PROCEDURE ExtractString (VAR str: ARRAY OF CHAR; len: SHORTINT; comp: ARRAY OF ARRAY OF CHAR; VAR m: SHORTINT);
VAR ccnt, wcnt, max: SHORTINT;
BEGIN
  wcnt:=0;
  LOOP
    ccnt:=0; IF len>0 THEN max:=len ELSE max:=SHORT(SHORT(LEN(comp,1)))-1 END;
    comp[wcnt][max]:=0X;  (* terminate string so we match shorter strings *)
    LOOP
      IF (str[ccnt]=0X) OR (CAP(str[ccnt])#CAP(comp[wcnt][ccnt])) THEN EXIT (* inner loop *) END;
      INC(ccnt);
      IF ccnt>max THEN EXIT (* inner loop *) END;
      IF comp[wcnt][ccnt]=0X THEN S.Delete(str, 0, ccnt); m:=wcnt+1; RETURN END (* exit both loops *)
    END;
    INC(wcnt);
    IF wcnt>=LEN(comp) THEN EXIT (* outer loop *) END
  END;
  m:=0  (* not found *)
END ExtractString;

PROCEDURE ExtractAMPM (VAR str: ARRAY OF CHAR; VAR hour: SHORTINT);
BEGIN
  IF (CAP(str[0])='A') & (CAP(str[1])='M') THEN S.Delete(str, 0, 2)
  ELSIF (CAP(str[0])='P') & (CAP(str[1])='M') THEN 
    S.Delete(str, 0, 2); INC(hour, 12);
    IF hour=24 THEN hour:=0 END
  END
END ExtractAMPM;

PROCEDURE ExtractZone (VAR str: ARRAY OF CHAR; VAR zone: INTEGER);
VAR
  TZAbbr: ARRAY 8 OF CHAR; ccnt: INTEGER; neg: BOOLEAN;
BEGIN
  TZAbbr:="GMT"; ccnt:=0;
  WHILE (TZAbbr[ccnt]#0X) & (TZAbbr[ccnt]=CAP(str[ccnt])) DO INC(ccnt) END; 
  IF TZAbbr[ccnt]=0X THEN (* found time zone *)
    S.Delete(str, 0, ccnt); 
    neg:=str[0]="+"; S.Delete(str, 0, 1);
    ExtractNumber(str, zone, 2); zone:=zone*60;
    ExtractNumber(str, ccnt, 2); INC(zone, ccnt);
    IF neg THEN zone:=-zone END
  ELSE zone:=SC.unknownZone
  END
END ExtractZone;


(* ------------------------------------------------------------- *)
(* String conversion functions *)

PROCEDURE (VAR c: Calendar) ToStr * (Pattern: ARRAY OF CHAR; 
                                     VAR DateStr: ARRAY OF CHAR);
(* 
     This function is similar to the `sprintf' function but the 
     conversion specifications that can appear in the format template 
     `Pattern' are specialized for printing components of the date and 
     time `c' according to the locale currently specified for time 
     conversion.

     Ordinary characters appearing in the `Pattern' are copied to the
     output string `DateStr'; this can include multibyte character
     sequences.  Conversion specifiers are introduced by a `%'
     character, and are replaced in the output string as follows:

    `%a'  The abbreviated weekday name according to the current locale.
    `%A'  The full weekday name according to the current locale.
    `%b'  The abbreviated month name according to the current locale.
    `%B'  The full month name according to the current locale.
    `%c'  The preferred date and time representation for the current
          locale.
    `%d'  The day of the month as a decimal number (range `01' to `31').
    `%D'  The day of the month as above but no leading zero.
    `%H'  The hour as a decimal number, using a 24-hour clock (range
          `00' to `23').
    `%I'  The hour as a decimal number, using a 12-hour clock (range
          `01' to `12').
     %i'  The hour as a decimal number, using a 12-hour clock with
          no leading zero.
    `%j'  The day of the year as a decimal number (range `001' to
          `366').
    `%m'  The month as a decimal number (range `01' to `12').
    `%M'  The minute as a decimal number.
    `%p'  Either `am' or `pm', according to the given time value; or the
          corresponding strings for the current locale.
    `%S'  The second as a decimal number.
    `%U'  The week number of the current year as a decimal number,
          starting with the first Sunday as the first day of the first
          week.
    `%W'  The week number of the current year as a decimal number,
          starting with the first Monday as the first day of the first
          week.
    `%w'  The day of the week as a decimal number, Sunday being `0'.
    `%x'  The preferred date representation for the current locale, but
          without the time.
    `%X'  The preferred time representation for the current locale, but
          with no date.
    `%y'  The year as a decimal number, but without a century (range
          `00' to `99').
    `%Y'  The year as a decimal number, including the century.
    `%Z'  The time zone or name or abbreviation (empty if the time zone
          can't be determined).
    `%%'  A literal `%' character.

   Therefore the pattern "%A, %D %b %Y" produces "Sunday, 12 Oct 1993" *)
CONST
  Lead0=2; 
VAR
  def, temp: ARRAY 256 OF CHAR;
  ccnt, rcnt: INTEGER;
  
  PROCEDURE AppendStr (str: ARRAY OF CHAR);
  VAR c: INTEGER;
  BEGIN
    c:=0; 
    WHILE str[c]#0X DO DateStr[rcnt]:=str[c]; INC(rcnt); INC(c) END
  END AppendStr;
  
  PROCEDURE AppendInt (int, zero: INTEGER);
  VAR
    Temp : ARRAY 10 OF CHAR;
  BEGIN
    IntStr.IntToStr(int, Temp);
    WHILE S.Length(Temp) < zero DO S.Insert("0", 0, Temp) END; 
    AppendStr(Temp)
  END AppendInt;
  
  PROCEDURE TwelveHour (hour: INTEGER) : INTEGER;
  BEGIN
    IF hour>12 THEN RETURN hour-12
    ELSIF hour=0 THEN RETURN 12
    ELSE RETURN hour
    END  
  END TwelveHour;

  PROCEDURE PutZone (zone: INTEGER);
  BEGIN
    IF zone<=SC.unknownZone THEN RETURN END;
    IF zone<0 THEN
      AppendStr("GMT+"); zone:=ABS(zone)
    ELSE AppendStr("GMT-")
    END;
    AppendInt(zone DIV 60, Lead0);
    AppendInt(zone MOD 60, Lead0)
  END PutZone;
  
BEGIN
  ccnt:=0; rcnt:=0;
  WHILE Pattern[ccnt]#0X DO
    IF (Pattern[ccnt]='%') & (Pattern[ccnt]#0X) THEN
      CASE Pattern[ccnt+1] OF
      | 'c': ASSERT(L.GetStr(L.defBothFormat, temp));
             c.ToStr(temp, def); AppendStr(def)
      | 'x': ASSERT(L.GetStr(L.defDateFormat, temp));
             c.ToStr(temp, def); AppendStr(def)
      | 'X': ASSERT(L.GetStr(L.defTimeFormat, temp));
             c.ToStr(temp, def); AppendStr(def)
      | '%': AppendStr("%")
      | 'D': AppendInt(c.day, 0)
      | 'd': AppendInt(c.day, Lead0)
      | 'j': AppendInt(c.DayOfYear(), Lead0+1)
      | 'B': c.MonthStr(FALSE, def); AppendStr(def)
      | 'b': c.MonthStr(TRUE, def); AppendStr(def)
      | 'm': AppendInt(c.month, Lead0)
      | 'U': AppendInt(c.WeekNumber(Sunday), Lead0)
      | 'W': AppendInt(c.WeekNumber(Monday), Lead0)
      | 'y': AppendInt(c.year MOD 100, Lead0)
      | 'Y': AppendInt(c.year, 0)
      | 'A': c.DayOfWeekStr(0, def); AppendStr(def)
      | 'a': c.DayOfWeekStr(3, def); AppendStr(def)
      | 'w': AppendInt(c.DayOfWeek(), 0)
      | 'Z': PutZone(c.zone)
      | 'H': AppendInt(c.hour, Lead0)
      | 'I': AppendInt(TwelveHour(c.hour), Lead0)
      | 'i': AppendInt(TwelveHour(c.hour), 0)
      | 'p': IF c.hour>=12 THEN AppendStr("PM") ELSE AppendStr("AM") END
      | 'M': AppendInt(c.minute, Lead0)
      | 'S': AppendInt(c.second, Lead0)
      | ELSE DateStr[rcnt]:='%'; DateStr[rcnt+1]:=Pattern[ccnt+1]
      END;
      INC(ccnt, 2)
    ELSE
      DateStr[rcnt]:=Pattern[ccnt]; INC(rcnt); INC(ccnt)
    END
  END;
  DateStr[rcnt]:=0X
END ToStr;

PROCEDURE InitFromStr * (VAR c: Calendar; DateStr: ARRAY OF CHAR; 
			 Pattern: ARRAY OF CHAR) : BOOLEAN;
(* Converts a day/time string `DateStr' into a date in `c' using
   the template contained in `Pattern'.  Pattern definitions are the
   same as in the DateToStr procedure.  This routine is not
   case-sensitive to months and days of the week.  An unspecified time
   or date is assumed as the zero time/date, respectively; after
   adjustment for the current time zone.  Incomplete times will assume
   the zero time for missing time elements.  Incomplete dates will
   assume the first day (if the day is omitted); the first month (if
   the month is omitted); and the zero year (if the year is omitted).
   Any other information such as the day of the week and the week
   number will be ignored.  The pattern "%A, %D %b %Y" is required to
   parse "Sunday, 12 Oct 1993".
   
   Successful parse of the `DateStr' according to the `Pattern' will
   return TRUE. *)

TYPE
  ShortString = ARRAY 32 OF CHAR;
   
VAR
  token: CHAR; lp, temp: ARRAY 80 OF CHAR; 
  inumb, decade: INTEGER; snumb: SHORTINT;
  err: BOOLEAN;
  d, m, s, min, h: SHORTINT;
  y, z: INTEGER;
  Months: ARRAY 12 OF ShortString;
  Days: ARRAY 7 OF ShortString;
  
  PROCEDURE NextToken;
  BEGIN
    IF lp[0]='%' THEN token:=lp[1]; S.Delete(lp, 0, 2)
    ELSIF lp#0X THEN 
      IF lp[0]#DateStr[0] THEN err:=TRUE END; (* mismatch *)
      S.Delete(lp, 0, 1); S.Delete(DateStr, 0, 1); token:=' '
    ELSE token:=0X
    END
  END NextToken;

  PROCEDURE InitMonths;
  VAR i: INTEGER;
  BEGIN
    FOR i:=0 TO 11 DO ASSERT(L.GetStr(L.months+i, Months[i])) END
  END InitMonths;

  PROCEDURE InitDays;
  VAR i: INTEGER;
  BEGIN
    FOR i:=0 TO 6 DO ASSERT(L.GetStr(L.daysOfWeek+i, Days[i])) END
  END InitDays;  

BEGIN
  (* basic approach is to look for either words or numbers depending
     on the expected parse token and then decode the received word as
     a month or a day of week string and numbers are decoded according
     to their parse token. *)
  Months[0]:=""; Days[0]:="";
  SC.GetClock(c);  (* get current date/time, time zone, daylight savings *)
  decade:=(c.year DIV 100) * 100;      (* set decade *)
  COPY(Pattern, lp); err:=FALSE;       (* assume no errors initially *)
  d:=1; m:=January; y:=1970;
  s:=0; min:=0; h:=0; z:=SC.unknownZone;
  REPEAT
    NextToken;
    CASE token OF
    | 'c': ASSERT(L.GetStr(L.defBothFormat, temp)); S.Insert(temp, 0, lp)
    | 'x': ASSERT(L.GetStr(L.defDateFormat, temp)); S.Insert(temp, 0, lp)
    | 'X': ASSERT(L.GetStr(L.defTimeFormat, temp)); S.Insert(temp, 0, lp)
    | 'D','d': ExtractNumber(DateStr, inumb, 2); d:=SHORT(inumb)
    | 'j': ExtractNumber(DateStr, inumb, 3) (* ignore *)
    | 'B': InitMonths; ExtractString(DateStr, 0, Months, m)
    | 'b': InitMonths; ExtractString(DateStr, 3, Months, m)
    | 'm': ExtractNumber(DateStr, inumb, 2); m:=SHORT(inumb)
    | 'U','W': ExtractNumber(DateStr, inumb, 2) (* ignore *)
    | 'y': ExtractNumber(DateStr, y, 2); INC(y, decade)
    | 'Y': ExtractNumber(DateStr, y, 4)
    | 'A': InitDays; ExtractString(DateStr, 0, Days, snumb) (* ignore *)
    | 'a': InitDays; ExtractString(DateStr, 3, Days, snumb) (* ignore *)
    | '%': S.Delete(DateStr, 0, 1)
    | 'Z': ExtractZone(DateStr, z)
    | 'H': ExtractNumber(DateStr, inumb, 2); h:=SHORT(inumb)
    | 'I','i': ExtractNumber(DateStr, inumb, 2); h:=SHORT(inumb)
    | 'p': ExtractAMPM(DateStr, h)
    | 'M': ExtractNumber(DateStr, inumb, 2); min:=SHORT(inumb)
    | 'S': ExtractNumber(DateStr, inumb, 2); s:=SHORT(inumb)
    | ELSE IF (token#0X)&(token#' ') THEN err:=TRUE END  (* unrecognized token *)
    END;
  UNTIL token=0X;
  
  (* attempt to fix up the date *)
  IF y<=0 THEN y:=1; err:=TRUE END;
  IF (m<January) OR (m>December) THEN m:=January; err:=TRUE END;
  IF (d<1) OR (d>31) THEN d:=1; err:=TRUE END;
  IF (h<0) OR (h>23) THEN h:=0; err:=TRUE END;
  IF (min<0) OR (min>59) THEN min:=0; err:=TRUE END;
  IF (s<0) OR (s>59) THEN s:=0; err:=TRUE END;

  (* finally set the date *)
  Init(c, d, m, y, h, min, s);
  c.zone:=z; c.summerTimeFlag:=SC.unknown;
  RETURN ~err
END InitFromStr;

PROCEDURE Test;
VAR s: ARRAY 80 OF CHAR; d: Calendar; i: INTEGER; p: LONGINT;
BEGIN
  SC.GetClock(d);
  d.ToStr("%A, %D %b, %Y", s); Out.String("Pattern: `%A, %D %b, %Y' = "); Out.String(s); Out.Ln;
  d.ToStr("Today is %c", s); Out.String("Pattern: `%c' = "); Out.String(s); Out.Ln; 
  IF ~InitFromStr(d, s, "Today is %c") THEN Out.String("Error detected..."); Out.Ln END;
  d.ToStr("Workweek = %W", s); Out.String(s); Out.Ln;
  d.ToStr("12-hour time = %i:%M:%S%p", s); Out.String(s); Out.Ln;
  d.ToStr("van Acken (weird pattern): %%c", s); Out.String(s); Out.Ln;  
  IF ~InitFromStr(d, s, "Full date: %A, %B %D, %Y, %j days from year start") THEN
    Out.String("Intentional error detected..."); Out.Ln
  END;
  SC.GetClock(d);
  d.ToStr("Full date: %A, %B %D, %Y, %j days from year start", s); Out.String(s); Out.Ln;
  IF ~InitFromStr(d, s, "Full date: %A, %B %D, %Y, %j days from year start") THEN
    Out.String("Error detected..."); Out.Ln
  END;
  Init(d, 29, December, 1995, 0, 0, 0);
  FOR i:=1 TO 400 DO
    d.ToStr("%a %b %D %Y, WW(Sun)=%U, WW(Mon)=%W", s); Out.String(s);
    p:=d.MoonPhase();
    IF p=0 THEN Out.String(" --> New Moon")
    ELSIF p=7 THEN Out.String(" --> First Quarter")
    ELSIF p=14 THEN Out.String(" --> Full Moon")
    ELSIF p=22 THEN Out.String(" --> Last Quarter")
    END;
    Out.Ln;
    d.Inc(1, 0)
  END;

  (* check Paschal & Easter dates *)
  (* Proper output should be:
	Paschal Full Moon: Wed Apr 3 1996; Easter: Sun Apr 7 1996
	Paschal Full Moon: Sun Mar 23 1997; Easter: Sun Mar 30 1997
	Paschal Full Moon: Sat Apr 11 1998; Easter: Sun Apr 12 1998
	Paschal Full Moon: Wed Mar 31 1999; Easter: Sun Apr 4 1999
	Paschal Full Moon: Wed Apr 19 2000; Easter: Sun Apr 23 2000
	Paschal Full Moon: Sun Apr 8 2001; Easter: Sun Apr 15 2001
	Paschal Full Moon: Thu Mar 28 2002; Easter: Sun Mar 31 2002
	Paschal Full Moon: Wed Apr 16 2003; Easter: Sun Apr 20 2003
	Paschal Full Moon: Mon Apr 5 2004; Easter: Sun Apr 11 2004
	Paschal Full Moon: Fri Mar 25 2005; Easter: Sun Mar 27 2005
	Paschal Full Moon: Thu Apr 13 2006; Easter: Sun Apr 16 2006
	Paschal Full Moon: Mon Apr 2 2007; Easter: Sun Apr 8 2007
	Paschal Full Moon: Sat Mar 22 2008; Easter: Sun Mar 23 2008
	Paschal Full Moon: Fri Apr 10 2009; Easter: Sun Apr 12 2009
	Paschal Full Moon: Tue Mar 30 2010; Easter: Sun Apr 4 2010
  *)
  FOR i:=1996 TO 2010 DO
    d.PaschalFullMoon(i);
    d.ToStr("Paschal Full Moon: %a %b %D %Y", s); Out.String(s);
    d.Inc(7-d.DayOfWeek(), 0);  (* advance to the next Sunday *)
    d.ToStr("; Easter: %a %b %D %Y", s); Out.String(s); Out.Ln
  END
END Test;


BEGIN 
  IF DEBUG THEN Test END
END Calendar.
