MODULE Locales;

(*
    Locales - localization information based on the Posix/C facilities.
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT C, Strings, Unix;

CONST
  (* Localization identifier definitions -- ids 0 to 99 are reserved *)
  (* NUMERIC category *)
  decimalPoint* = 0;      (* decimal point separator *)
  thousandsSep* = 1;      (* digit group separator to left of dp *)
  grouping* = 2;          (* string groups to left of dp *)

  (* MONETARY category *)
  intCurrencySymbol* = 3; (* international currency symbol *)
  currencySymbol* = 4;    (* local currency symbol *)
  monDecimalPoint* = 5;   (* monetary decimal point separator *)
  monThousandsSep* = 6;   (* monetary digit group separator to left of dp *)
  monGrouping* = 7;       (* monetary string groups to left of dp *)
  positiveSign* = 8;      (* positive monetary quantities *)
  negativeSign* = 9;      (* negative monetary quantities *)
  intFracDigits* = 10;    (* international fractional digits to right of dp *)
  fracDigits* = 11;       (* local fractional digits to right of dp *)
  pCSPrecedes* = 12;      (* positive monetary symbol precedes *)
  pSepBySpace* = 13;      (* positive monetary symbol separated by space *)
  nCSPrecedes* = 14;      (* negative monetary symbol precedes *)
  nSepBySpace* = 15;      (* negative monetary symbol separated by space *)
  pSignPosn* = 16;        (* positive number sign position *)
  nSignPosn* = 17;        (* negative number sign position *)

  (* TIME category *)
  daysOfWeek* = 18;       (* day of week strings -- 18-Sunday, 19-Monday, etc. *)
  months* = daysOfWeek+7; (* month strings -- 25-January, 26-Februrary, etc. *)
  defTimeFormat* = months+12;       (* default time format *)
  defDateFormat* = defTimeFormat+1; (* default date format *)
  defBothFormat* = defDateFormat+1; (* default time/date format *)
  
  startUserID* = 100;

  (* any integer item which is not specified *)
  unspecified* = MAX(SHORTINT);

VAR
  base: Unix.tmDesc;   (* base time unit *)
  
  (* Localization category definitions *)
  ALL- : LONGINT;      (* all possible parameters *)
  COLLATE- : LONGINT;  (* string collation *)
  CTYPE- : LONGINT;    (* character conversion *) 
  MONETARY- : LONGINT; (* money formatting parameters *)
  NUMERIC- : LONGINT;  (* number formatting parameters *)
  TIME- : LONGINT;     (* time/date parameters *)  


PROCEDURE RecoverPattern(VAR str: ARRAY OF CHAR);
(* This function takes a formatted string from the base date
   and attempts to recover the pattern which generated the
   output string. We proceed by converting all parts of the
   base date into output strings and then match these strings
   against the `str' and replace the matches with the used
   template. *)

  PROCEDURE ReplacePattern(pattern: ARRAY OF CHAR);
  VAR 
    ret: LONGINT; pos: INTEGER; 
    src: ARRAY 132 OF CHAR;
    found: BOOLEAN;
  BEGIN
    ret:=Unix.strftime(src, 132, pattern, base);
    Strings.FindNext(src, str, 0, found, pos);
    IF found THEN (* replace the found pattern *)
      Strings.Delete(str, pos, Strings.Length(src));
      Strings.Insert(pattern, pos, str)
    END
  END ReplacePattern;

BEGIN
  (* check and replace all possible patterns *)
  (* if anyone has any better way of doing this, let me know *)
  ReplacePattern("%A"); ReplacePattern("%a"); ReplacePattern("%B");
  ReplacePattern("%b"); ReplacePattern("%d"); ReplacePattern("%H");
  ReplacePattern("%I"); ReplacePattern("%j"); ReplacePattern("%m");
  ReplacePattern("%M"); ReplacePattern("%p"); ReplacePattern("%S");
  ReplacePattern("%U"); ReplacePattern("%W"); ReplacePattern("%w");
  ReplacePattern("%Y"); ReplacePattern("%y")
END RecoverPattern;


PROCEDURE GetStr * (id: LONGINT; VAR ls: ARRAY OF CHAR) : BOOLEAN;
(* Return the localization string in `ls' which is associated with
   the `id'.  Return FALSE if the localization `id' is invalid, 
   undefined, or not a string. *)
VAR
  loc: Unix.lconv; str: C.charPtr1d; tp: Unix.tmDesc; ret: LONGINT;
BEGIN
  (* acquire the information from the C localization *)
  IF id<daysOfWeek THEN
    (* use the C localization information *)
    loc:=Unix.localeconv();
    CASE id OF
    | decimalPoint:      str:=loc.decimal_point
    | thousandsSep:      str:=loc.thousands_sep
    | grouping:          str:=loc.grouping
    | intCurrencySymbol: str:=loc.int_curr_symbol
    | currencySymbol:    str:=loc.currency_symbol
    | monDecimalPoint:   str:=loc.mon_decimal_point
    | monThousandsSep:   str:=loc.mon_thousands_sep
    | monGrouping:       str:=loc.mon_grouping
    | positiveSign:      str:=loc.positive_sign
    | negativeSign:      str:=loc.negative_sign
    | ELSE RETURN FALSE
    END;
    COPY(str^, ls)
  ELSIF id<months THEN
    (* use the strftime routine to get the day strings *)
    tp:=base; tp.tm_wday:=id-daysOfWeek; INC(tp.tm_mday, tp.tm_wday);
    ret:=Unix.strftime(ls, LEN(ls), "%A", tp)
  ELSIF id<defTimeFormat THEN
    (* use the strftime routine to get the month strings *)
    tp:=base; tp.tm_mon:=id-months;
    ret:=Unix.strftime(ls, LEN(ls), "%B", tp)
  ELSIF id=defTimeFormat THEN
    ret:=Unix.strftime(ls, LEN(ls), "%X", base);
    RecoverPattern(ls) 
  ELSIF id=defDateFormat THEN 
    ret:=Unix.strftime(ls, LEN(ls), "%x", base);
    RecoverPattern(ls) 
  ELSIF id=defBothFormat THEN 
    ret:=Unix.strftime(ls, LEN(ls), "%c", base);
    RecoverPattern(ls) 
  ELSE
    RETURN FALSE
  END;
  RETURN TRUE
END GetStr;

PROCEDURE GetInt * (id: LONGINT) : LONGINT;
(* Return the localization integer which is associated with the `id'.  
   Return MIN(LONGINT) if the localization `id' is invalid, undefined, 
   or not an integer. *)
VAR
  loc: Unix.lconv;
BEGIN
  loc:=Unix.localeconv();
  CASE id OF
  | intFracDigits: RETURN loc.int_frac_digits
  | fracDigits:    RETURN loc.frac_digits
  | pCSPrecedes:   RETURN loc.p_cs_precedes
  | pSepBySpace:   RETURN loc.p_sep_by_space
  | nCSPrecedes:   RETURN loc.n_cs_precedes
  | nSepBySpace:   RETURN loc.n_sep_by_space
  | pSignPosn:     RETURN loc.p_sign_posn
  | nSignPosn:     RETURN loc.n_sign_posn
  | ELSE           RETURN MIN(LONGINT) 
  END
END GetInt;

PROCEDURE Set * (category: LONGINT; name: ARRAY OF CHAR; VAR old: ARRAY OF CHAR);
(* Set a new locale called `name' for the `category' items.  Return the
   previous locale name in `old'. NOTE: Unless the following call is
   made in a user program, all locales will use the "C" defaults and
   environment variables will be ignored:
   
                Locales.Set(Locales.ALL, "", old)
 *)
VAR
  str: C.charPtr1d;
BEGIN
  str:=Unix.setlocale(category, name);
  IF str=NIL THEN COPY("", old)
  ELSE COPY(str^, old)
  END
END Set;

BEGIN
  (* initialize a base date = Sunday, 22 Dec 1996, 16:02:03 *)
  base.tm_sec:=3; base.tm_min:=2; base.tm_hour:=16;
  base.tm_mday:=22; base.tm_mon:=11; base.tm_year:=96; base.tm_wday:=0;
  
  (* these definitions are valid for SunOS 4 *)
  ALL:=0;      (* all possible parameters, same on HP-UX and BeBox *)
  COLLATE:=6;  (* string collation *)
  CTYPE:=1;    (* character conversion *) 
  MONETARY:=4; (* money formatting parameters *)
  NUMERIC:=2;  (* number formatting parameters *)
  TIME:=3;     (* time/date parameters *)  
END Locales.
