MODULE SysClock;
  
(*
    SysClock - facilities for accessing a system clock that records the 
               date and time of day.
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT JD:=JulianDay, Unix;
  
CONST
  secsPerDay     = 24*60*60;
  secsPerHalfDay = secsPerDay DIV 2;
  TZActive       = TRUE;   (* default for Unix systems *)

  maxSecondParts * = 999; (* Most systems have just millisecond accuracy *) 
  
  zoneMin * = -780;       (* time zone minimum minutes *)
  zoneMax * =  720;       (* time zone maximum minutes *)

  localTime * = MIN(INTEGER);  (* time zone is inactive & time is local *)
  unknownZone * = localTime+1; (* time zone is unknown *)
 
  (* daylight savings mode values *)
  unknown * = -1;   (* current daylight savings status is unknown *)
  inactive * = 0;   (* daylight savings adjustments are not in effect *)
  active * = 1;     (* daylight savings adjustments are being used *)
    
TYPE
  (* The DateTime type is a system-independent time format whose fields
     are defined as follows:
     
          Year     > 0
          Month    = 1 .. 12
          Day      = 1 .. 31
          Hour     = 0 .. 23
          Min      = 0 .. 59
          Sec      = 0 .. 59
          Fraction = 0 .. maxSecondParts
          Zone     = -780 .. 720
          
     Note: We adopt the Gregorian date convention as required on the
     original adoption date on 15 Oct 1582.  The British did not
     adopt this calendar until 3 Sep 1752.  The UseGregorian flag
     will switch between using the Gregorian calendar when TRUE and
     using the Julian calendar when FALSE.  When using the calendar,
     the dates will automatically switch to using the Julian calendar
     if a date is requested which is before the Gregorian calendar
	 adoption date.
  *)     
  DateTime * =
    RECORD
      year*:           INTEGER;
      month*:          SHORTINT;
      day*:            SHORTINT;
      hour*:           SHORTINT;
      minute*:         SHORTINT;
      second*:         SHORTINT;
      summerTimeFlag*: SHORTINT; (* daylight savings mode (see above) *)      
      fractions*:      INTEGER;  (* parts of a second in milliseconds *)
      zone*:           INTEGER;  (* Time zone differential factor which 
                                    is the number of minutes to add to 
                                    local time to obtain GMT or is set
                                    to localTime when time zones are
                                    inactive. *)
    END;

VAR
  startJD, endJD: LONGINT;  (* system start and end dates as Julian days *) 
   
PROCEDURE CanGetClock * (): BOOLEAN;
  (* Returns TRUE if a system clock can be read; FALSE otherwise *)
VAR t: Unix.timeval;
BEGIN
  RETURN Unix.gettimeofday(t, Unix.NULL)=0
END CanGetClock;
   
PROCEDURE CanSetClock * (): BOOLEAN;
  (* Returns TRUE if a system clock can be set; FALSE otherwise *)
BEGIN
  RETURN Unix.getuid()=0
END CanSetClock;

PROCEDURE DaysPerMonth (mon: SHORTINT; year: INTEGER) : SHORTINT;
VAR
  day: SHORTINT;
BEGIN
  CASE mon OF
  |  1, 3, 5, 7, 8, 10, 12: RETURN 31
  |  2: JD.DaysToDate(JD.DateToDays(28, mon, year)+1, day, mon, year);
        IF mon#2 THEN RETURN 28 ELSE RETURN 29 END
  |  ELSE RETURN 30
  END
END DaysPerMonth;

PROCEDURE IsValidDateTime * (d: DateTime): BOOLEAN;
(* Returns TRUE if the value of `d' represents a valid date and time; 
   FALSE otherwise *)
BEGIN
  RETURN (d.month>0) & (d.month<=12) &
         (d.day>0) & (d.day<=DaysPerMonth(d.month, d.year)) &
         ((d.zone<=unknownZone) OR (d.zone>=zoneMin) & (d.zone<=zoneMax)) &
		 (d.hour>=0) & (d.hour<60) &
         (d.minute>=0) & (d.minute<60) &
         (d.second>=0) & (d.second<60) &
         (d.fractions>=0) & (d.fractions<=maxSecondParts)
END IsValidDateTime;

PROCEDURE DateToSecs (d: DateTime) : LONGINT;
(* Returns number of seconds since 1 Jan 1970 or -1 on
   invalid input dates.  Seconds are in GMT.
   Pre: 1 Jan 1970, 0:00:00<=d<=1 Jan 2106, 0:00:00; d.zone is valid 
   Post: Seconds as an unsigned 32-bit number *)
VAR 
  deltaDays, secs: LONGINT;
BEGIN
  (* difference between ref. date and `d' *)
  deltaDays:=JD.DateToDays(d.day, d.month, d.year)-startJD;
  IF deltaDays<0 THEN RETURN -1 END;
   
  (* calculate the seconds from the days--using unsigned 32-bits *)
  IF d.zone<zoneMin THEN d.zone:=0 END;
  secs:=deltaDays*secsPerHalfDay+(d.hour*60+d.zone+d.minute)*30;
  IF secs>(MAX(LONGINT)-d.second) DIV 2 THEN INC(secs, MIN(LONGINT)) END;
  RETURN secs*2+d.second
END DateToSecs;

PROCEDURE SecsToDate (s: LONGINT; tz: INTEGER; dst: SHORTINT; VAR d: DateTime);
(* Converts the number of seconds (GMT) since system start 
   date to a valid date/time in `d' in the time zone `tz' which is in
   minutes.  The `tz' is meant to be added to local time to obtain GMT.
   `dst' is TRUE when daylight savings time is in effect.  
   Note: `s' is treated as an unsigned 32-bit integer.
   Pre: 0<=s<100000000;
   Post: `d' contains date relative to 1 Jan 1970 *)
VAR 
  lobit: BOOLEAN; days: LONGINT;
BEGIN
  (* check if we are using the sign bit *)
  IF s<0 THEN 
    lobit:=ODD(s);
    s:=(s DIV 2) - MIN(LONGINT);
    days:=s DIV secsPerHalfDay;
    DEC(s, days*secsPerHalfDay);
    IF lobit THEN INC(s) END;
    days:=2*days; s:=2*s
  ELSE days:=s DIV secsPerDay; s:=s MOD secsPerDay 
  END;
 
  (* determine the date *)
  d.summerTimeFlag:=dst; d.zone:=tz; d.fractions:=0;
  d.second:=SHORT(SHORT(s MOD 60)); s:=s DIV 60;
  IF tz>=zoneMin THEN DEC(s, LONG(tz)) END;
  d.minute:=SHORT(SHORT(s MOD 60));
  d.hour:=SHORT(SHORT(s DIV 60));
  JD.DaysToDate(days+startJD, d.day, d.month, d.year)
END SecsToDate;
 
PROCEDURE GetClock * (VAR userData: DateTime);
(* If possible, assigns system date and time of day to userData.
   Error returns 1 Jan 1970. *)
VAR t: LONGINT; tm: Unix.tm;
BEGIN 
  IF CanGetClock() THEN 
    t:=Unix.time(Unix.NULL);   (* could also use gettimeofday *)
    tm:=Unix.localtime(t);
    userData.zone:=localTime;
    userData.summerTimeFlag:=unknown;
    userData.second:=SHORT(SHORT(tm.tm_sec));
    userData.minute:=SHORT(SHORT(tm.tm_min));
    userData.hour:=SHORT(SHORT(tm.tm_hour));
    userData.day:=SHORT(SHORT(tm.tm_mday));
    userData.month:=SHORT(SHORT(tm.tm_mon+1));
    userData.year:=SHORT(tm.tm_year+1900);	
    IF TZActive THEN
      (* determine the time zone offset GMT-Local *)
      userData.zone:=SHORT((t-DateToSecs(userData)) DIV 60);
      userData.summerTimeFlag:=SHORT(SHORT(tm.tm_isdst))
    END;
    userData.fractions:=0
  ELSE SecsToDate(0, unknownZone, unknown, userData)
  END
END GetClock;
   
PROCEDURE SetClock * (userData: DateTime);
(* If possible, sets the system clock to the values of userData *)
VAR t: Unix.timeval; stat: LONGINT;
BEGIN
  stat:=JD.DateToDays(userData.day, userData.month, userData.year);
  IF IsValidDateTime(userData) & (stat>=startJD) & (stat<=endJD) THEN 
    t.tv_sec:=DateToSecs(userData);
    t.tv_usec:=userData.fractions*(maxSecondParts+1);
    stat:=Unix.settimeofday(t, Unix.NULL)
  END
END SetClock;

BEGIN
  (* start/end dates for seconds conversion *)
  startJD:=JD.DateToDays(1, 1, 1970);
  endJD:=JD.DateToDays(1, 1, 2106)
END SysClock.
