(*	$Id: TextRider.Mod,v 1.2 1997/02/09 07:05:53 acken Exp $	*)
MODULE TextRider;

(*
    TextRider -  Text-based input/output of Oberon variables.       
    Copyright (C) 1997 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT
  Ascii, Channel, CharClass, Strings, LRealStr, RealStr, IntStr, 
  LRealConv, ConvTypes;

CONST
  done* = Channel.done;
  invalidFormat* = Channel.invalidFormat;

  (* Scanner types *)
  char*=0; string*=1; int*=2; real*=3; bool*=4; set*=5;
  tab*=6; line*=7; name*=8; eot*=9; invalid*=10;
  
  (* Writer options *)
  noBuffering*=0;        (* when set output is not buffered.  This allows
                            interactive output prompts to appear as soon as
                            they are written *)

  (* Reader/Scanner options *)
  returnCtrlChars*=0;    (* when set returns EOL & TAB characters; otherwise they
                            are treated like whitespace *)
                           
  (* additional Scanner options *)
  interpretBools*=1;     (* when set returns truth values of the strings "TRUE"
                            and "FALSE"; otherwise returns the strings *)
  interpretSets*=2;      (* when set returns a set value of the string set
                            representation; otherwise returns the brace and
                            comma characters, and the numbers individually *)
                            
  defReaderOptions*  = {};
  defWriterOptions*  = {};
  defScannerOptions* = defReaderOptions + {interpretBools, interpretSets};

TYPE
  Reader* = POINTER TO ReaderDesc;
  ReaderDesc* = RECORD
    opt-: SET;                  (* current reader options (see above) *)
    full: BOOLEAN;              (* pch holds a valid character *)
    pch: CHAR;                  (* previous read character *)
    byteReader: Channel.Reader;
    base-: Channel.Channel;
  END;

  Writer* = POINTER TO WriterDesc;
  WriterDesc* = RECORD 
    opt-: SET;                  (* current writer options (see above) *)
    byteWriter: Channel.Writer;
    base-: Channel.Channel;
  END;

  String* = ARRAY 256 OF CHAR;
  Scanner* = POINTER TO ScannerDesc;
  ScannerDesc* = RECORD
    r: Reader;       (* attached reader *)
    base-: Channel.Channel;
    opt-: SET;       (* current scanner options (see above) *)
    type-: INTEGER;  (* scanned token type (see above) *)
    pos-: LONGINT;   (* position of current token in the channel *) 
    lines-: LONGINT; (* total lines scanned *)
    int-: LONGINT;   (* integer from channel *)
    real-: LONGREAL; (* real from channel *)
    char-: CHAR;     (* char from channel *)
    string-: String; (* string from channel *)
    set-: SET;       (* set from channel *)
    bool-: BOOLEAN;  (* boolean from channel *)
  END;

(* Reader methods 
   ------------------------------------------------------------------------ *)

(* The following methods read a value of the given type from the current
   position in the TextReader. Iff the value is invalid for its type, 
   'r.Res' returns 'invalidFormat'.
  *)
  
PROCEDURE (r: Reader) Pos* () : LONGINT;
  VAR
    pos: LONGINT;
  BEGIN
    pos:=r.byteReader.Pos();
    IF (pos>0) & r.full THEN DEC(pos) END;
    RETURN pos
  END Pos;

PROCEDURE (r: Reader) Available* () : LONGINT;
  VAR
    avail: LONGINT;
  BEGIN
    avail:=r.byteReader.Available();
    IF (avail<0) & r.full THEN RETURN 1
    ELSIF r.full THEN RETURN avail+1
    ELSE RETURN avail
    END
  END Available;

PROCEDURE (r: Reader) SetPos* (newPos: LONGINT);
  BEGIN
    r.byteReader.SetPos(newPos);
    r.full:=FALSE                 (* character is not valid *)
  END SetPos;

PROCEDURE (r: Reader) Res* () : INTEGER;
  BEGIN
    RETURN r.byteReader.res
  END Res;

PROCEDURE (r: Reader) SetOpts* (opts: SET);
(* Set the reader options `r.opt' which are defined above. *)
  BEGIN
    r.opt:=opts
  END SetOpts;

PROCEDURE (r: Reader) ReadChar* (VAR ch: CHAR);
  BEGIN
    IF r.full THEN ch:=r.pch; r.full:=FALSE
    ELSE r.byteReader.ReadByte(ch); (* SIZE(s.BYTE) = SIZE(CHAR) *)
    END
  END ReadChar;
  
PROCEDURE (r: Reader) UngetChar(ch: CHAR);
  BEGIN
    ASSERT(~r.full);  (* only allow single-level unget *)
    r.pch:=ch; r.full:=TRUE 
  END UngetChar;

PROCEDURE (r: Reader) Eol*(): BOOLEAN;
(* Return TRUE if the character at the current position is the 
   system-dependent end-of-line char or the last character has
   been read
 *)
  VAR
    ch: CHAR;
  BEGIN
    r.ReadChar(ch); 
    IF r.byteReader.res#done THEN RETURN TRUE END;
    r.UngetChar(ch);
    RETURN CharClass.IsEOL(ch)
  END Eol;
 
PROCEDURE (r: Reader) SkipBlanks (VAR ch: CHAR);
  VAR
    allowCtrl: BOOLEAN;
  BEGIN
    allowCtrl:=returnCtrlChars IN r.opt;
    REPEAT
      r.ReadChar(ch)
    UNTIL (r.byteReader.res#done) OR (ch>" ") OR (allowCtrl & ((ch=CharClass.eol) OR (ch=Ascii.ht)))
  END SkipBlanks;

PROCEDURE (r: Reader) ReadLn*;
  VAR 
    ch: CHAR;
  BEGIN
    REPEAT
      r. ReadChar (ch)
    UNTIL (r. byteReader. res # done) OR (ch = CharClass.eol)
  END ReadLn;

PROCEDURE (r: Reader) Check(pat: ARRAY OF CHAR) : BOOLEAN;
  VAR
    pos: INTEGER;
    ch: CHAR;
  BEGIN
    pos:=0; 
    WHILE pat[pos]#0X DO
      r.ReadChar(ch);
      IF (r.byteReader.res#done) OR (ch#pat[pos]) THEN RETURN FALSE END;
      INC(pos)
    END;
    RETURN TRUE
  END Check;

PROCEDURE (r: Reader) ReadBool* (VAR bool: BOOLEAN);
(* Pre: input=[whitespace] ["TRUE"|"FALSE"]; Post: bool=TRUE iff input="TRUE"
   and bool=FALSE iff input="FALSE"; undefined otherwise *)
  VAR
    ch: CHAR;
  BEGIN
    r.SkipBlanks(ch);
    IF (r.byteReader.res=done) & (ch="T") THEN
      IF r.Check("RUE") THEN bool:=TRUE
      ELSE r.byteReader.SetRes(invalidFormat)
      END
    ELSIF (r.byteReader.res=done) & (ch="F") THEN
      IF r.Check("ALSE") THEN bool:=FALSE
      ELSE r.byteReader.SetRes(invalidFormat)
      END
    ELSE r.byteReader.SetRes(invalidFormat)
    END
  END ReadBool;

PROCEDURE (r: Reader) ReadString* (VAR s: ARRAY OF CHAR);
(* Pre: input = [whitespace] '"' {char} '"' | [whitespace] "'" {char} "'"
   Illegal chars terminate with invalidFormat.
 *)
  VAR
    cnt: INTEGER; quote: CHAR;
  BEGIN
    r.SkipBlanks(quote);

    (* check for leading quote *)
    IF (quote#'"') & (quote#"'") THEN
      r.byteReader.SetRes(invalidFormat); RETURN
    END;

    (* read in the characters *)
    cnt:=-1;
    REPEAT 
      INC(cnt); r.ReadChar(s[cnt]);
      IF s[cnt]<" " THEN  r.byteReader.SetRes(invalidFormat); RETURN END 
    UNTIL (cnt=LEN(s)-1) OR (s[cnt]=quote) OR (r.byteReader.res#done);
    s[cnt]:=0X   (* terminate string *)
  END ReadString;

PROCEDURE (r: Reader) ReadLine* (VAR s: ARRAY OF CHAR);
(* Reads characters until an end of line character is encountered, or the 
   array `s' is full.  CAUTION: If reading multiple lines of input and an
   integer, real, etc. has just been read, the channel may be positioned
   at a eol character and this method will return an empty string. *)
  VAR
    cnt: INTEGER;
  BEGIN
    (* read in the characters *)
    cnt:=-1;
    REPEAT 
      INC(cnt); r.ReadChar(s[cnt]);
    UNTIL (cnt=LEN(s)-1) OR (s[cnt]=CharClass.eol);
    IF (s[cnt] # CharClass.eol) THEN
      r. UngetChar (s[cnt])
    END;
    s[cnt]:=0X (* terminate string *)
  END ReadLine;

PROCEDURE (r: Reader) ReadIdentifier* (VAR s: ARRAY OF CHAR);
(* Pre: input = [whitespace] letter {letter | digit}
 *)
  VAR
    cnt: INTEGER;
  BEGIN
    r.SkipBlanks(s[0]); cnt:=0;
    IF ~CharClass.IsLetter(s[0]) THEN 
      r.byteReader.SetRes(invalidFormat); RETURN
    END;

    (* read in the rest of the identifier *)
    REPEAT INC(cnt); r.ReadChar(s[cnt]) 
    UNTIL (cnt=LEN(s)-1) OR (r.byteReader.res#done) OR 
          (~CharClass.IsLetter(s[cnt]) & ~CharClass.IsNumeric(s[cnt]));
    
    (* unget the last character *)
    IF r.byteReader.res=done THEN r.UngetChar(s[cnt]) END;
    s[cnt]:=0X (* terminate string *)
  END ReadIdentifier;
 
PROCEDURE (r: Reader) StateScan (VAR str: ARRAY OF CHAR; class: ConvTypes.ScanClass; 
                                 state: ConvTypes.ScanState);
  VAR
    pos: INTEGER; ch: CHAR;
  BEGIN
    pos:=1;
    LOOP
      IF r.byteReader.res#done THEN EXIT END;
      r.ReadChar(ch);
      state.p(ch, class, state);
      IF class=ConvTypes.valid THEN str[pos]:=ch; INC(pos)
      ELSE EXIT
      END
    END;  

    (* unget the last non-digit *)
    r.UngetChar(ch); str[pos]:=0X;
  END StateScan;

PROCEDURE (r: Reader) HexDigit (ch: CHAR) : BOOLEAN;
  BEGIN
    RETURN ((ch>="0") & (ch<="9")) OR ((ch>="A") & (ch<="F"))
  END HexDigit;

PROCEDURE (r: Reader) HexToInt (str: ARRAY OF CHAR; VAR lint: LONGINT);
(* Returns the long integer constant `lint' in the string `str' according 
   to the format:
     IntConst = digit {hexdigit}

   Note: 80000000H-FFFFFFFFH are valid inputs which map to the negative
   integers. *)
  CONST
    BASE   = 16;
    MAXPAT = 8000000H;
  VAR
    d, pos: INTEGER;
  BEGIN    
    (* accept the hexadecimal input number *)
    lint:=0; pos:=0;
    LOOP
      (* read a digit *)
      d:=ORD(str[pos]);
      IF d=0 THEN EXIT
      ELSIF CharClass.IsNumeric(CHR(d)) THEN DEC(d, ORD("0"))
      ELSE (* A-F *) DEC(d, ORD("A")-10)
      END;
      
      (* check for overflow and adjustment *)
      IF (lint>=MAXPAT*2) OR (lint<0) THEN 
        r.byteReader.SetRes(invalidFormat); RETURN (* overflow *)
      ELSIF (lint>=MAXPAT) & (d>=8) THEN 
        DEC(lint, MAXPAT*2)                        (* adjustment *)
      END;

      (* build up the number *)
      lint:=BASE*lint+d; 
      INC(pos)
    END 
  END HexToInt;

PROCEDURE (r: Reader) ReadLInt* (VAR lint: LONGINT);
(* Returns the long integer constant n at the current position according to the
   format:
           IntConst = [whitespace] ["+"|"-"] digit {digit}        
*)
  VAR
    str: ARRAY 64 OF CHAR; 
    pos: INTEGER; res: SHORTINT;
  BEGIN 
    r.SkipBlanks(str[0]);
   
    (* check for the sign -- if any *)
    IF (str[0]="+") OR (str[0]="-") THEN r.ReadChar(str[1]); pos:=1
    ELSE pos:=0
    END;
    
    (* check whether the first digit is valid *)
    IF ~CharClass.IsNumeric(str[pos]) THEN
      r.byteReader.SetRes(invalidFormat); RETURN
    END;

    (* accumulate the digits *)
    WHILE (r.byteReader.res=done) & CharClass.IsNumeric(str[pos]) DO
      INC(pos); r.ReadChar(str[pos])
    END;

   (* convert to an integer *)
    r.UngetChar(str[pos]); str[pos]:=0X; 
    IntStr.StrToInt(str, lint, res);
   
    (* set errors -- if needed *)
    IF res#IntStr.strAllRight THEN r.byteReader.SetRes(invalidFormat) END
  END ReadLInt;
  
PROCEDURE (r: Reader) ReadHex* (VAR lint: LONGINT);
(* Returns the long integer constant n at the current position according to the
   format:
     IntConst = [whitespace] digit {hexdigit}
                where hexDigit = "0".."9" | "A".."F" 
   Note: Numbers in the range 80000000H-FFFFFFFFH are read in as negative
   numbers.  If numbers like 80H-FFH are to be interpreted as negatives for
   SHORTINTs then neg = lint-100H or 8000H-FFFFH are to be interpreted as 
   negatives for INTEGERs then neg = lint-10000H.
*)
  VAR
    str: ARRAY 64 OF CHAR; 
    pos: INTEGER;
  BEGIN 
    r.SkipBlanks(str[0]); pos:=0;

    (* ensure first character is numeric *)
    IF ~CharClass.IsNumeric(str[0]) THEN
      r.byteReader.SetRes(invalidFormat); RETURN (* illegal digit *)
    END;

    (* accumulate the digits *)
    WHILE (r.byteReader.res=done) & (r.HexDigit(str[pos])) DO
      INC(pos); r.ReadChar(str[pos])
    END;

    (* convert to an integer *)
    r.UngetChar(str[pos]); str[pos]:=0X; 
    r.HexToInt(str, lint)  (* HexToInt sets errors internally *)     
  END ReadHex;

PROCEDURE (r: Reader) ReadInt * (VAR int: INTEGER);
(* as ReadLInt *)
  VAR
    lint: LONGINT;
  BEGIN
    r.ReadLInt(lint);
    IF (lint>MAX(INTEGER)) OR (lint<MIN(INTEGER)) THEN
      r.byteReader.SetRes(invalidFormat)
    ELSE int:=SHORT(lint)
    END
  END ReadInt;

PROCEDURE (r: Reader) ReadSInt * (VAR sint: SHORTINT);
(* as ReadLInt *)
  VAR
    lint: LONGINT;
  BEGIN
    r.ReadLInt(lint);
    IF (lint>MAX(SHORTINT)) OR (lint<MIN(SHORTINT)) THEN
      r.byteReader.SetRes(invalidFormat)
    ELSE sint:=SHORT(SHORT(lint))
    END
  END ReadSInt;

PROCEDURE (r: Reader) ReadLReal* (VAR lreal: LONGREAL);
(* Returns the long real constant `lreal' at the current position according to the 
   format:
     LongRealConst = [whitespace] ["+" | "-"] digit {digit} ["." {digit} [exponent]]
   where exponent = ("E" | "D") ("+" | "-") digit {digit}
*) 
  VAR
    str: ARRAY 1024 OF CHAR; res: SHORTINT;
    class: ConvTypes.ScanClass; state: ConvTypes.ScanState;
  BEGIN    
    (* use state machine to ensure valid input *)
    r.ReadChar(str[0]);
    LRealConv.ScanReal(str[0], class, state);
    r.StateScan(str, class, state);    
      
    (* convert the real string *)
    LRealStr.StrToReal(str, lreal, res);
    
    (* set errors -- if needed *)
    IF res#IntStr.strAllRight THEN
      r.byteReader.SetRes(invalidFormat)
    END
  END ReadLReal;

PROCEDURE (r: Reader) ReadReal* (VAR real: REAL);
(* as ReadLReal *) 
  VAR
    n: LONGREAL;
  BEGIN
    r.ReadLReal(n);
    IF (n<MIN(REAL)) OR (n>MAX(REAL)) THEN
      r.byteReader.SetRes(invalidFormat)
    ELSE
      real:=SHORT(n)
    END
  END ReadReal;

PROCEDURE (r: Reader) ReadSet* (VAR s: SET);
(* Read a set described in mathematical set notation into 's'.
   Pre: "{Element, ..., Element}"; Post: s={Element, ..., Element}
   where Element = number [".." number] and 0 <= number <= 31 *)
  VAR
    ch: CHAR; lo, hi: SHORTINT;
    
    PROCEDURE err;
    BEGIN
      r.byteReader.SetRes(invalidFormat);
    END err;
    
  BEGIN
    (* ensure a valid start *)
    r.SkipBlanks(ch);
    IF ch#"{" THEN err; RETURN END;

    (* read in the set *)
    s:={}; lo:=-1;
    WHILE (r.byteReader.res=done) & (ch#"}") DO
      r.ReadChar(ch);
      CASE ch OF
      | "0".."9":
             r.UngetChar(ch); r.ReadSInt(lo); hi:=lo;
             IF (hi<MIN(SET)) OR (hi>MAX(SET)) THEN err; RETURN END                  
      | ".": IF r.Check(".") THEN r.ReadSInt(hi);      
               IF (lo>hi) OR (hi<MIN(SET)) OR (hi>MAX(SET)) THEN err; RETURN END
             ELSE err; RETURN
             END;
      | " ": (* ignore blanks *)
      | ",", "}": 
             IF lo#-1 THEN s:=s+{lo..hi} END
      | ELSE err; RETURN
      END
    END;
    
    IF r.byteReader.res#done THEN err END
  END ReadSet;
 

(* Scanner methods 
   ------------------------------------------------------------------------ *)
   
PROCEDURE (s: Scanner) Get (VAR ch: CHAR);
(* local Read routine to capture eols & eot *)
  BEGIN
    s.r.ReadChar(ch);
    IF s.r.byteReader.res#done THEN ch:=0X    
    ELSIF ch=CharClass.eol THEN INC(s.lines) 
    END
  END Get;

PROCEDURE (s: Scanner) Skip* (VAR ch: CHAR);
(* Skip white space as defined by the scanner options.  Return the first
   non-white space character. *)
  VAR
    allowCtrl: BOOLEAN;
  BEGIN
    allowCtrl:=returnCtrlChars IN s.opt;
    REPEAT
      s.Get(ch)
    UNTIL (ch=0X) OR (ch>" ") OR (allowCtrl & ((ch=CharClass.eol) OR (ch=Ascii.ht)))
  END Skip;

PROCEDURE (s: Scanner) Pos* () : LONGINT;
(* Position of the look-ahead character *)
  BEGIN
    RETURN s.r.Pos()
  END Pos;

PROCEDURE (s: Scanner) Available* () : LONGINT;
(* Position of the look-ahead character *)
  BEGIN
    RETURN s.r.Available()
  END Available;

PROCEDURE (s: Scanner) SetPos* (pos: LONGINT);
  BEGIN
    s.r.SetPos(pos)
  END SetPos;

(* Scan for the next token *)
PROCEDURE (s: Scanner) Scan*;
  VAR
    ch: CHAR; str: String; pos: INTEGER; res: SHORTINT;
  BEGIN
    s.Skip(ch); s.r.UngetChar(ch); (* skip leading blanks *)
    s.pos:=s.Pos();
    CASE ch OF
    | '"', "'":
           s.r.ReadString(s.string); s.type:=string
    | "a".."z", "A".."Z": 
           s.r.ReadIdentifier(s.string);
           IF interpretBools IN s.opt THEN 
             IF s.string="TRUE" THEN s.bool:=TRUE; s.type:=bool
             ELSIF s.string="FALSE" THEN s.bool:=FALSE; s.type:=bool
             ELSE s.type:=name
             END
           ELSE s.type:=name
           END
    | "+", "-", "0".."9": (* integer or real *)
           (* get first digit/sign *)
           pos:=0; s.Get(str[0]);
           
           (* read leading digits *)
           REPEAT INC(pos); s.Get(str[pos])
           UNTIL ~s.r.HexDigit(str[pos]);
           
           (* check for reals or hexadecimals *)
           IF str[pos]="." THEN  (* real number *)
             s.type:=real;
           
             (* read trailing digits *)
             REPEAT INC(pos); s.Get(str[pos])
             UNTIL ~CharClass.IsNumeric(str[pos]);
             
             (* read the exponent *)
             IF (str[pos]="E") OR (str[pos]="D") THEN
               INC(pos); s.Get(str[pos]);   (* get sign *)        
               REPEAT INC(pos); s.Get(str[pos])
               UNTIL ~CharClass.IsNumeric(str[pos])
             ELSE s.r.UngetChar(str[pos])
             END;
             
             (* convert to real *)
             str[pos]:=0X;
             LRealStr.StrToReal(str, s.real, res);
    
             (* set errors -- if needed *)
             IF res#IntStr.strAllRight THEN
               s.r.byteReader.SetRes(invalidFormat)
             END             
           ELSIF str[pos]="H" THEN (* hexadecimal integer *)
             s.type:=int; str[pos]:=0X; s.r.HexToInt(str, s.int)
           ELSE (* just an integer *)
             s.type:=int; s.r.UngetChar(str[pos]); str[pos]:=0X;
             IntStr.StrToInt(str, s.int, res);
             IF res#IntStr.strAllRight THEN s.r.byteReader.SetRes(invalidFormat) END             
           END
    | "{": IF interpretSets IN s.opt THEN s.r.ReadSet(s.set); s.type:=set
           ELSE s.Get(s.char); s.type:=char
           END
    | 0X:  s.type:=eot
    | ELSE s.Get(s.char); s.type:=char;
           IF s.char=CharClass.eol THEN s.type:=line
           ELSIF s.char=Ascii.ht THEN s.type:=tab
           END
    END
  END Scan;

(* Set the scanner options `s.opt' which are defined above. *)
PROCEDURE (s: Scanner) SetOpts* (opts: SET);
  BEGIN
    s.opt:=opts;
    s.r.opt:=opts*{returnCtrlChars}  (* adjust the reader options as well *)
  END SetOpts;


(* Writer methods 
   ------------------------------------------------------------------------ *)

(* The following write methods write the value as a string to the 
   underlying Channel.
 *)
 
PROCEDURE (w: Writer) Pos* () : LONGINT;
  BEGIN
    RETURN w.byteWriter.Pos()
  END Pos;

PROCEDURE (w: Writer) SetPos* (newPos: LONGINT);
  BEGIN
    w.byteWriter.SetPos(newPos)
  END SetPos;

PROCEDURE (w: Writer) Res* () : INTEGER;
  BEGIN
    RETURN w.byteWriter.res
  END Res;

PROCEDURE (w: Writer) SetOpts* (opts: SET);
(* Set the writer options `w.opt' which are defined above. *)
  BEGIN
    w.opt:=opts
  END SetOpts;

PROCEDURE (w: Writer) Flush*;
  BEGIN
    w.byteWriter.base.Flush
  END Flush;

(* The terminating 0X is not written *)
PROCEDURE (w: Writer) WriteString*(s: ARRAY OF CHAR);
  BEGIN
    w. byteWriter. WriteBytes (s, 0, Strings.Length (s));
    IF noBuffering IN w.opt THEN w.Flush END
  END WriteString;

PROCEDURE (w: Writer) WriteBool*(bool: BOOLEAN);
  BEGIN
    IF bool THEN w. WriteString ("TRUE")
    ELSE w. WriteString ("FALSE")
    END
  END WriteBool;
  
PROCEDURE (w: Writer) WriteChar*(ch: CHAR);
  BEGIN
    w.byteWriter.WriteByte(ch);
    IF noBuffering IN w.opt THEN w.Flush END    
  END WriteChar;
  
PROCEDURE (w: Writer) WritePad (n: LONGINT);
  BEGIN
    (* output padding *)
    WHILE n>0 DO w.WriteChar(" "); DEC(n) END  
  END WritePad;
  
(* Convert 'sint' to a string of at least 'n' chars and write it to the
   underlying channel. If 'n' is too small it will be extended. If 'n'
   is greater then nessesary spaces will be added after the number, i.e.
   it is left justified. *)
PROCEDURE (w: Writer) WriteLInt*(lint: LONGINT; n: LONGINT);
  VAR
    val: ARRAY 16 OF CHAR;
  BEGIN
    (* convert to a string *)
    IntStr.IntToStr(lint, val);
    
    (* output any required padding *)
    w.WritePad(n-Strings.Length(val));
    
    (* output the string *)
    w.WriteString(val)
  END WriteLInt;
  
PROCEDURE (w: Writer) WriteSInt* (sint: SHORTINT; n: LONGINT);
  BEGIN
    w.WriteLInt(sint, n)
  END WriteSInt;

PROCEDURE (w: Writer) WriteInt* (int: INTEGER; n: LONGINT);
  BEGIN
    w.WriteLInt(int, n)
  END WriteInt;
 
PROCEDURE (w: Writer) WriteHexDigits(VAR n: LONGINT; digits: LONGINT);
  CONST
    BASE=16;
  VAR
    dig: LONGINT;
  BEGIN
    (* output padding digits *)
    WHILE digits>8 DO 
      IF n<0 THEN w.WriteChar("F") ELSE w.WriteChar("0") END;
      DEC(digits)
    END;

    (* output the actual number *)
    WHILE digits>0 DO
      DEC(digits);
      dig := ASH(n, -4*digits) MOD BASE;
      IF dig<=9 THEN w.WriteChar(CHR(ORD("0") + dig)) 
      ELSE w.WriteChar(CHR(ORD("A") - 10 + dig)) 
      END
    END
  END WriteHexDigits;
  
(* Write `lint' as a heximal number using `d' digits. 
   If `d' <= 0 then `lint' is written using 8 digits. *)
PROCEDURE (w: Writer) WriteHex* (lint: LONGINT; d: LONGINT);
  BEGIN
    IF d<=0 THEN d:=8 END;
    w.WriteHexDigits(lint, d)
  END WriteHex;

PROCEDURE (w: Writer) WriteLReal*(lreal: LONGREAL; n, k: LONGINT);
 (* The call to WriteLongReal(lreal, n, k) shall write to the underlying
    channel a formatted string corresponding to the value of `lreal' in
    floating-point form.  A sign shall be included only for negative
    values.  One significant digit shall be included in the whole number
    part.  The signed exponent part shall be included only if the exponent
    value is not 0.  If the value of `k' is greater than 0, that
    number of significant digits shall be included, otherwise an
    implementation-defined number of significant digits shall be
    included.  The decimal point shall not be included if there are no
    significant digits in the fractional part.  The complete formatted
    number will be right-justified in a field of width `n'.  If 'n' is 
    too small it will be extended.
    
    For example: (n=9)
    
    value:     3923009     39.23009     0.0003923009
      k
      1             4E+6        4E+1         4E-4 
      2           3.9E+6      3.9E+1       3.9E-4
      5        3.9230E+6   3.9230E+1    3.9230E-4
 *)
  VAR
    val: ARRAY 128 OF CHAR;
  BEGIN
    (* convert to a string *)
    LRealStr.RealToFloat(lreal, SHORT(k), val);
    
    (* output any required padding *)
    w.WritePad(n-Strings.Length(val));

    (* output the string *)
    w.WriteString(val);
  END WriteLReal;

PROCEDURE (w: Writer) WriteReal*(real: REAL; n, k: LONGINT);
(* As WriteLReal *)
  VAR
    val: ARRAY 128 OF CHAR;
  BEGIN
    (* convert to a string *)
    RealStr.RealToFloat(real, SHORT(k), val);
    
    (* output any required padding *)
    w.WritePad(n-Strings.Length(val));

    (* output the string *)
    w.WriteString(val)
  END WriteReal;

PROCEDURE (w: Writer) WriteLRealFix*(VAR lreal: LONGREAL; n, k: LONGINT);
 (* 
    The call WriteLRealFix(lreal, n, k) shall output to the underlying
    channel the formatted string corresponding to the value of `lreal' in
    fixed-point form.  A sign shall be included only for negative values.
    At least one digit shall be included in the whole number part.  The
    value shall be rounded to the given value of `k' relative to the
    decimal point.  The decimal point shall be suppressed if `k' is
    less than 0. The complete formatted number will be right-justified 
    in a field of width `n'.  If 'n' is too small it will be extended.
    
    For example: (n=12)
    
    value:        3923009         3.923009   0.0003923009
      k
     -5             3920000             0          0 
     -2             3923010             0          0
     -1             3923009             4          0 
      0            3923009.            4.         0. 
      1           3923009.0           3.9        0.0
      4        3923009.0000        3.9230     0.0004       
 *)
  VAR
    val: ARRAY 128 OF CHAR;
  BEGIN
    (* convert to a string *)
    LRealStr.RealToFixed(lreal, SHORT(k), val);
    
    (* output any required padding *)
    w.WritePad(n-Strings.Length(val));
    
    (* output the string *)
    w.WriteString(val)
  END WriteLRealFix;

PROCEDURE (w: Writer) WriteRealFix*(real: REAL; n, k: LONGINT);
(* As WriteLongRealFix *)
  VAR
    val: ARRAY 128 OF CHAR;
  BEGIN
    (* convert to a string *)
    RealStr.RealToFixed(real, SHORT(k), val);
    
    (* output any required padding *)
    w.WritePad(n-Strings.Length(val));
    
    (* output the string *)
    w.WriteString(val)
  END WriteRealFix;

PROCEDURE (w: Writer) WriteLRealEng*(VAR lreal: LONGREAL; n, k: LONGINT);
 (* 
    Converts the value of real to floating-point string form, with 
    `k' significant figures, and writes the resultant value,
    right-justified in a field of width `n' to the underlying channel.
    If 'n' is too small it will be extended. The number is scaled with 
    one to three digits in the whole number part and with an exponent 
    that is a multiple of three.
     
    For example: n=9
    
    value:     3923009     39.23009   0.0003923009
      k
      1             4E+6        40         400E-6 
      2           3.9E+6        39         390E-6
      5        3.9230E+6    39.230      392.30E-6     
  *)
  VAR
    val: ARRAY 128 OF CHAR;
  BEGIN
    (* convert to a string *)
    LRealStr.RealToEng(lreal, SHORT(k), val);
    
    (* output any required padding *)
    w.WritePad(n-Strings.Length(val));
    
    (* output the string *)
    w.WriteString(val)
  END WriteLRealEng;

PROCEDURE (w: Writer) WriteRealEng*(real: REAL; n, k: LONGINT);
(* As WriteLRealEng *)
  VAR
    val: ARRAY 128 OF CHAR;
  BEGIN
    (* convert to a string *)
    RealStr.RealToEng(real, SHORT(k), val);
    
    (* output any required padding *)
    w.WritePad(n-Strings.Length(val));
    
    (* output the string *)
    w.WriteString(val)
  END WriteRealEng;
  
PROCEDURE (w: Writer) WriteSet*(s: SET);
(* Write 's' in mathematical set notation.
   Pre: x = {Element, ..., Element}; Post: write "{Element, ..., Element}" 
   where Element = number [".." number] and 0 <= number <= 31 *)
  VAR
    bit, lo: SHORTINT; addComma: BOOLEAN;
  BEGIN
    w.WriteChar("{"); bit:=0; addComma:=FALSE;
    WHILE bit<=MAX(SET) DO      
      IF bit IN s THEN
        lo:=bit;
        WHILE (bit<MAX(SET)) & (bit+1 IN s) DO INC(bit) END; (* check for runs *)
        
        (* output the set element(s) *)
        IF addComma THEN w.WriteString(", ") ELSE addComma:=TRUE END;
        w.WriteInt(lo, 0);
        IF lo<bit THEN
          w.WriteString(".."); w.WriteInt(bit, 0)
        END
      END;
      INC(bit)
    END;
    w.WriteChar("}");
  END WriteSet;
 
PROCEDURE (w: Writer) WriteLn*;
(* Write a newline *)
  BEGIN
    w.WriteChar(CharClass.eol)
  END WriteLn;


(* Reader Procedures
   ------------------------------------------------------------------------ *)

PROCEDURE ConnectReader*(ch: Channel.Channel): Reader;
  VAR
    r: Reader; t: Channel.Reader;
  BEGIN
    t:=ch.NewReader();
    IF t=NIL THEN RETURN NIL END;
    NEW(r);
    r.full:=FALSE;   (* no previous characters *)
    r.opt:=defReaderOptions;
    r.byteReader:=t;
    r. base := ch;
    RETURN r
  END ConnectReader;
  
(* Writer Procedures
   ------------------------------------------------------------------------ *)

PROCEDURE ConnectWriter*(ch: Channel.Channel): Writer;
  VAR
    w: Writer; t: Channel.Writer;
  BEGIN
    t:=ch.NewWriter();
    IF t=NIL THEN RETURN NIL END;
    NEW(w);
    w.opt:=defWriterOptions;
    w.byteWriter:=t;
    w. base := ch;
    RETURN w
  END ConnectWriter;

(* Scanner Procedures
   ------------------------------------------------------------------------ *)

PROCEDURE ConnectScanner*(ch: Channel.Channel): Scanner;
  VAR
    s: Scanner; t: Reader;
  BEGIN
    t:=ConnectReader(ch);
    IF t=NIL THEN RETURN NIL END;
    NEW(s);
    s.r:=t;
    s.opt:=defScannerOptions;
    s.type:=invalid;
    s.lines:=0;
    s. base := ch;
    RETURN s
  END ConnectScanner;
  

END TextRider.

